var gc = gc || {};
gc.databind = gc.databind || {};

(function()
{
    var FIFOCommandResponseQueue = function(name)
    {
        this.name = name;
        this.first = this.last = this;
    };

    FIFOCommandResponseQueue.prototype.addCommand = function(command, sequenceNumber)
    {
        var deferred = Q.defer();
        
        this.last.next =
        {
            seqNo : sequenceNumber,
            command: command,
            deferred: deferred
        };
        this.last = this.last.next;
        
        return deferred.promise;
    };
    
    FIFOCommandResponseQueue.prototype.addResponse = function(response, command, sequenceNumber, isError)
    {
        var step = -1;
        while (this.first !== this.last && step < 0)
        {
            var cur = this.first.next;
            step = sequenceNumber === undefined ? 0 : cur.seqNo - sequenceNumber;
            if (step < -100) 
            {
                step += 255;
            }
            else if (step > 100)
            {
                step -= 255;
            }
                
            if (step < 0)
            {
                cur.deferred.reject(this.name + ' error: missing response for command sequence #' + cur.deferred.seqNo);
                this.first = cur;
            }
            else if (step === 0)
            {
                if (cur.command !== command)
                {
                    cur.deferred.reject(this.name + ' error: Command Mismatch.  Expected ' + cur.command + ', but received ' + command);
                }
                else if (isError)
                {
                    cur.deferred.reject(response);
                }
                else 
                {
                    cur.deferred.resolve(response);
                }
                this.first = cur;
            }
        }
    };
    
    FIFOCommandResponseQueue.prototype.clearAll = function()
    {
        while (this.first !== this.last)
        {
            this.first = this.first.next;
            
            this.first.deferred.reject('Skipping response from ' + this.name + ' due to reset operation');
        }
        this.first = this.last = this;
    };
    
    FIFOCommandResponseQueue.prototype.isEmpty = function()
    {
        return this.first === this.last;
    };
    
    /** 
     * Abstract class for command/response, message based, packet codecs.  This class manages messages
     * sent and received from a controller/firmware.  This class provides helper methods for adding 
     * commands to a queue waiting for responses from the controller/firmware and associating responses with
     * the appropriate command.  Messages can optionally have command id's as well as sequence numbers to aid in
     * matching up responses to commands.  If not provided, all commands are expected to have responses, and then must
     * be processed in order.  If sequence numbers are provided, commands will be matched up by sequence number, but 
     * they still must be processed in order.  Out of order responses, with sequence numbers, are not supported. 
     *    
     * @constructor
     * @implements gc.databind.IPollingPacketCodec
     */
    gc.databind.AbstractMessageBasedCodec = function(name)
    {
        this.name = name;
        this.commandQueue = new FIFOCommandResponseQueue(name);
        this.disconnect();
    };

    gc.databind.AbstractMessageBasedCodec.prototype = new gc.databind.IPollingPacketCodec();

    /**
     * Default implementation of the connect method that does nothing.  If initialization is  
     * required, derived implementations should implement their own version of this method and 
     * return a promise to indicate when the connection is completed. 
     */
    gc.databind.AbstractMessageBasedCodec.prototype.connect = function()
    {
        return Q();
    };

    /**
     * Implementation of the disconnect method.  This method clears the commad/response queue on disconnect.
     * If you override this method, be sure to call this base method in addition to other cleanup activities.  
     */
    gc.databind.AbstractMessageBasedCodec.prototype.disconnect = function()
    {
        this.commandQueue.clearAll();
    };
    
    var logPacket = function(name, message,  buffer, len)
    {
        gc.console.log(name, function() 
        {
            len = len || buffer.length;
            for(var i = 0; i < len; i++)
            {
                message = message + buffer[i].toString(16) + ', ';
            }
            return message;
        });
    };

    /**
     * Implementation of the IPacketCodec.encode() method.  This implementation logs the encoded packet using 
     * gc.console.log() method for debug purposes before forwarding the packet to the next 
     * encoder in the packet encoding chain.  The derived implementation should call this 
     * method to send packets to the target.
     * 
     * @param {function} target - the target encoder to call with the encoded packet for transmission.
     * @param {object} value - the packet for transmission.
     */
    gc.databind.AbstractMessageBasedCodec.prototype.encode = function(target, value)
    {
        logPacket(this.name, 'send    ', value);
        target(value);
    };
    
    /**
     * Implementation of the IPacketCodec.decode() method.  This implementation logs the incoming packet 
     * to gc.console.log().  The derived implementation should override this method to implement the 
     * necessary decoding of incomping packet, and it should call this base method first to perform the 
     * logging function, before decoding the incoming and passing it along to the target decoder/model.
     * 
     * @param {function} target - the target decoder to call with the decoded packet received.
     * @param {object} value - the packet received for decoding.
     */
    gc.databind.AbstractMessageBasedCodec.prototype.decode = function(target, value)
    {
        logPacket(this.name, 'receive ', value);
    };

    /**
     * Helper method to add a command to the command/response queue.  This method returns a promise that can be used
     * to process the response received by this method.  Only messages that require response handling need be 
     * added to the command/response queue.  Messages that are sent that do not require handling may be omitted. 
     * 
     * @param [Number|String] command - the specific command sent that requires a response.
     * @param [Number] sequence - the sequence number of the packet that requires a response.
     */
    gc.databind.AbstractMessageBasedCodec.prototype.addCommand = function(command, sequenceNumber)
    {
        return this.commandQueue.addCommand(command, sequenceNumber);
    };
    
    /**
     * Helper method to process responses in command/response queue.  This method will find the appropriate 
     * command in the command/response queue and reject any pending commands that did not receive a response, based
     * on the optional sequence numbers provided. 
     * 
     * @param {Array|String} response - the raw response message received.
     * @param [Number|String] command - the specific command that this response is for.
     * @param [Number] sequence - the sequence number of the packet received, if there is one.
     */
    gc.databind.AbstractMessageBasedCodec.prototype.addResponse = function(response, toCommand, sequenceNumber)
    {
        return this.commandQueue.addResponse(response, toCommand, sequenceNumber);
    };
    
    /**
     * Helper method to process error responses in command/response queue.  This method will find the appropriate 
     * command in the command/response queue and reject it.  Any pending commands that did not receive a response, 
	 * based on the optional sequence numbers provided, will also be rejected. 
     * 
     * @param {String} response - the error message for the command.
     * @param [Number|String] command - the specific command that this error is for.
     * @param [Number] sequence - the sequence number of the packet received, if there is one.
     */
    gc.databind.AbstractMessageBasedCodec.prototype.addErrorResponse = function(response, toCommand, sequenceNumber)
    {
        return this.commandQueue.addResponse(response, toCommand, sequenceNumber, true);
    };
    
    /**
     * Method to determine if the target is still connected.  The ti-transport-usb calls this api to
     * determine if tere has been a loss of connection if it has not received any data in a while.  This method returns
     * a promise that either resolves if there are no commands in the queue expecting a response; otherwise, if returns fails, indicating that
     * there are commands that have not been responded to.
     * 
     * @returns promise that resolves if still connected, or fails if the connection is lost with an error message.
     */
    gc.databind.AbstractMessageBasedCodec.prototype.isConnected = function()
    {
        if (!this.commandQueue.isEmpty())
        {
            return Q.reject('No response from ' + this.name + ' controller');
        }
        return Q(true);
    };

}());